## ----setup, include = FALSE---------------------------------------------------
library(fsbc)
library(magrittr)
library(Biostrings)
library(stringr)

## -----------------------------------------------------------------------------
DS <- readDNAStringSet("data/sample.fst")
DS

## -----------------------------------------------------------------------------
lmin <- 5
lmax <- 10

# Calculate frequency.
DS.freq    <- fsbc_calc_freq(DS)
# Get probability of nucleobases.
BR         <- fsbc_get_base_ratio(DS)
# Select subsequences.
DF.subseq  <- fsbc_search_subseq(DS.freq, DS.freq@metadata$freq, symbols = BR)
# Generate clusters with selected subsequences. 
L.cluster  <- fsbc_seq_cluster(rownames(DF.subseq), DS.freq)
# Add cluster ID to the cluster object. 
DS.cluster <- fsbc_label_cluster(L.cluster, DS.freq)

## ---- fig.width = 15, fig.height = 8, fig.cap = "Frequency of oligonucleotide sequences"----
plot(DS.freq@metadata$rank, DS.freq@metadata$freq, pch = 19,
		 xlab = "Ranking", ylab = "Frequency")

## ---- fig.width = 15, fig.height = 8, fig.cap = "Nucleobase ratio"------------
barplot(as.vector(BR), col = 1:4, names = names(BR), ylim = c(0,1),
	main = "Nucleobase ratio")

## -----------------------------------------------------------------------------
DF.subseq[1:12,]

## ---- fig.width = 15, fig.height = 8, fig.cap = "Distribution of $Z$-score and $Z^*$-score"----
par(mfrow = c(1,2))
boxplot(DF.subseq$Z  ~ DF.subseq$L, main = "Z-score")
boxplot(DF.subseq$ZZ ~ DF.subseq$L, main = "Z*-score")

## -----------------------------------------------------------------------------
all <- sum(sapply(lmin:lmax, function(i) 4^i))
nrow(DF.subseq) / all

## ---- fig.width = 15, fig.height = 8------------------------------------------
freq <- DS.cluster@metadata$freq # Frequency of sequence
cid  <- DS.cluster@metadata$cluster.id # Cluster ranking
plot(cid, freq, pch = 19, ylab = "Frequency", xlab = "Cluster ranking", log = "x")

## ----fig.width = 15, fig.height = 8, fig.cap = "Clustering Results"-----------
cluster.div <- sapply(L.cluster, length)
l <- factor(nchar(names(cluster.div)))
bp <- barplot(cluster.div, names = 1:length(cluster.div), col = l, ylab = "Number of unique sequences")
legend("topright", title = " Length of strings ", levels(l), col = 1:nlevels(l), pch = 15)
text(bp[,1], cluster.div, names(cluster.div), pos = 3, xpd = T)

## ---- fig.width = 15, fig.height = 8------------------------------------------
cluster.freq <- by(DS.cluster@metadata$freq, DS.cluster@metadata$cluster.id, sum)
l <- factor(nchar(names(L.cluster)))
bp <- barplot(cluster.freq, names = 1:length(cluster.freq), col = l, ylab = "Number of unique sequences")
legend("topright", title = " Length of strings ", levels(l), col = 1:nlevels(l), pch = 15)
text(bp[,1], cluster.freq, names(L.cluster), pos = 3, xpd = T)

## -----------------------------------------------------------------------------
sapply(L.cluster, head, 1) %>% head(., 5)
# $CGCAGGTGA
#   A DNAStringSet instance of length 1
#     width seq                                               names               
# [1]    30 TGGGGTGGGCGCAGGTGAGGGGGTGGGGGT                    R:1.F:319
# 
# $TGCTA
#   A DNAStringSet instance of length 1
#     width seq                                               names               
# [1]    30 GCGGGGGGTGCTAGGGCGGAGGTGGGCGTT                    R:2.F:278
# 
# $CAGGTGA
#   A DNAStringSet instance of length 1
#     width seq                                               names               
# [1]    30 TGGGGTGGGCACAGGTGAGGGGGTGGGGGT                    R:6.F:8
# 
# $CTAGGGCGGA
#   A DNAStringSet instance of length 1
#     width seq                                               names               
# [1]    30 GCGGGGGGCGCTAGGGCGGAGGTGGGCGTT                    R:4.F:16
# 
# $CGCAGGT
#   A DNAStringSet instance of length 1
#     width seq                                               names               
# [1]    30 TGGGGTGGGCGCAGGTAAGGGGATGGGGGT                    R:66.F:1
# 

